/*------------------------------------------------
--                 LAZZE-GETOBZ                 --
--------------------------------------------------
-- Filename      : map.c
-- Description   : Map data handling 
--                 
--
-- Author        : Lars Englund
--                 d99las@dtek.chalmers.se
--
--                 Tobias Gedell
--                 d99getob@dtek.chalmers.se
--
-- Created       : 020831
-- Last update   : 020902
------------------------------------------------*/


#include <stdlib.h>
#include <string.h>

#include "error.h"
#include "game.h"
#include "map.h"


int mapWidth;
int mapHeight;

int yoff[1000];             // Y offsets for map arrays
char *gameMap;              // Holds the gamemap
unsigned char *visitMap;    // Holds the number of times each square has been visited
int *distMap;               // Holds the distances
char *moveMap;              // Holds temporary visit information for the robots move function
int *scoreMap;              // Holds score data
int *packetMap;             // Holds packet data
char *destMap;              // Holds destination data


void allocMaps(int width, int height)
{
    int n,m;
    
    mapWidth=width;
    mapHeight=height;

    // Calculate y-offset
    for( n=0; n<height; n++ ) {
        yoff[n] = width*n;
    }
    
    // Allocate game map data
    if( !(gameMap = (char*) malloc( sizeof(char) * width * height )) ) {
        ERROR( "Could not allocate game map data" );
    }

    // Allocate visits map data
    if( !(visitMap = (unsigned char*) malloc( sizeof(char) * width * height )) ) {
        ERROR( "Could not allocate visits map data" );
    }

    // Allocate dist map data
    if( !(distMap = (int*) malloc( sizeof(int) * width * height )) ) {
        ERROR( "Could not allocate dist map data" );
    }

    // Allocate moves map data
    if( !(moveMap = (char*) malloc( sizeof(char) * width * height )) ) {
        ERROR( "Could not allocate moves map data" );
    }

    // Allocate score map data
    if( !(scoreMap = (int*) malloc( sizeof(int) * width * height )) ) {
        ERROR( "Could not allocate score map data" );
    }

    // Allocate packet map data
    if( !(packetMap = (int*) malloc( sizeof(int) * width * height )) ) {
        ERROR( "Could not allocate packet map data" );
    }

    // Allocate dest map data
    if( !(destMap = (char*) malloc( sizeof(char) * width * height )) ) {
        ERROR( "Could not allocate dest map data" );
    }
    
    memset(gameMap, 0, sizeof(char) * width * height);
    memset(visitMap, 0, sizeof(char) * width * height);
    memset(distMap, 0, sizeof(int) * width * height);
    memset(moveMap, 0, sizeof(char) * width * height);
    memset(scoreMap, 0, sizeof(int) * width * height);
    memset(packetMap, 0, sizeof(int) * width * height);
    memset(destMap, 0, sizeof(char) * width * height);
}


void clearMoveMap()
{
    memset(moveMap, 0, sizeof(char) * mapWidth * mapHeight);
}

void clearVisitMap()
{
    memset(visitMap, 0, sizeof(char) * mapWidth * mapHeight);
}

void clearScoreMap()
{
    memset(scoreMap, 0, sizeof(int) * mapWidth * mapHeight);
}

void freeMaps()
{
    // Free map data
    free( gameMap );
    free( visitMap );
    free( distMap );
    free( moveMap );
    free( scoreMap );
    free( packetMap );
    free( destMap );
}


// Game map interaction
char getGamePoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( WALL );     // Return wall if we try to index outside the map
    }
    return( gameMap[yoff[y] + x] );
}

void putGamePoint( int x, int y, char point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        gameMap[yoff[y] + x] = point;
    }
}


// Visits map interaction
unsigned char getVisitPoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( 255 );
    }
    return( visitMap[yoff[y] + x] );
}

void putVisitPoint( int x, int y, unsigned char point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        visitMap[yoff[y] + x] = point;
    }
}

void incVisitPoint(int x, int y)
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        visitMap[yoff[y] + x]++;
    }
}

void decVisitPoint(int x, int y)
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) )
    {
        int value=visitMap[yoff[y]+x]-1;
        
        value=(value<1)?0:value;

        visitMap[yoff[y]+x]=value;
    }
}


// Dest map interaction
int getDistPoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( 30000 );     // Return alot if we try to index outside the map
    }
    return( distMap[yoff[y] + x] );
}

void putDistPoint( int x, int y, int point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        distMap[yoff[y] + x] = point;
    }
}


// Moves map interaction
char getMovePoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( 1 );     // Return 1 if we try to index outside the map
    }
    return( moveMap[yoff[y] + x] );
}

void putMovePoint( int x, int y, char point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        moveMap[yoff[y] + x] = point;
    }
}


// Score map interaction
int getScorePoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( 0 );     // Return 0 if we try to index outside the map
    }
    return( scoreMap[yoff[y] + x] );
}

void putScorePoint( int x, int y, int point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        scoreMap[yoff[y] + x] = point;
    }
}


// Packet map interaction
int getPacketPoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( 0 );     // Return 0 if we try to index outside the map
    }
    return( packetMap[yoff[y] + x] );
}

void putPacketPoint( int x, int y, int point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        packetMap[yoff[y] + x] = point;
    }
}

void incPacketPoint(int x, int y)
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        packetMap[yoff[y] + x]++;
    }
}

void decPacketPoint(int x, int y)
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) )
    {
        int value=packetMap[yoff[y]+x]-1;
        
        value=(value<1)?0:value;

        packetMap[yoff[y]+x]=value;
    }
}

// Dest map interaction
char getDestPoint( int x, int y )
{
    if( y<0 || y>=mapHeight || x<0 || x>=mapWidth ) {
        return( 1 );     // Return wall if we try to index outside the map
    }
    return( destMap[yoff[y] + x] );
}

void putDestPoint( int x, int y, char point )
{
    if( !(y<0 || y>=mapHeight || x<0 || x>=mapWidth) ) {
        destMap[yoff[y] + x] = point;
    }
}
