/*------------------------------------------------
--                 LAZZE-GETOBZ                 --
--------------------------------------------------
-- Filename      : robot.c
-- Description   : Robot functionality 
--                 
--
-- Author        : Lars Englund
--                 d99las@dtek.chalmers.se
--
--                 Tobias Gedell
--                 d99getob@dtek.chalmers.se
--
-- Created       : 020831
-- Last update   : 020902
------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>

#include "error.h"
#include "game.h"
#include "gdebug.h"
#include "robot.h"
#include "map.h"
#include "network.h"


#define ROBOT_FOV           12
#define ROBOT_MAX_MOVES     10000
#define SCORE_BREAK         0

#define SCOREHOLD           97

#define MAX(a, b)           ((a>b)?a:b)

typedef struct _moveStruct
{
    int x;
    int y;
    int score;
    int count;
    struct _moveStruct *children[4];
} moveStruct;


// Robot information
int robotX;
int robotY;
int robotId;
int robotCapacity;
int robotMoney;
int robotLookingForPackets=1;

int lastXMove=LEFT;
int lastYMove=DOWN;

//- Move lists
moveStruct *moves1[1000];
moveStruct *moves2[1000];
moveStruct **pmoves1=&moves1[0];
moveStruct **pmoves2=&moves2[0];

//- Move heap
moveStruct moveHeap[ROBOT_MAX_MOVES];
moveStruct *pmoveHeap = &moveHeap[0];

//- Move pointers
moveStruct **move;
moveStruct **nextMove;


//- Move counters
int numMoves;
int numNextMoves;
int numHeapMoves;

//- Move diffs
int moveX[]={0, 1, 0, -1};
int moveY[]={1, 0, -1, 0};

void dispDebugInfo();
void dispDebugInfoMove(char *str, moveStruct *move);


packetStruct *heaviestPacketToPick;


int score(moveStruct *mv)
{
    int score;
    char point=getGamePoint(mv->x, mv->y);
    
    //- Stop if wall or water
    if(point == WALL) return -1;
    if(point == WATER)  return -1;


    //- Stop for close enemies
    //- AVOIDANCE
    if(mv->count<3 && robotLookingForPackets==0)
    {
        playerStruct *player = pfirstPlayer;
        while( player )
        {
            if( player->x == mv->x && player->y == mv->y )
                return -1;
            player=player->next;
        }
    }    
    
    score=0;

    //- Check if packet destination
    if(getDestPoint(mv->x, mv->y)!=0)
    {
        //- Go through all packets we are carrying
        int ggr;
        
        for(ggr=0;ggr<robotNumPackets;ggr++)
        {
            packetStruct *packet=robotPackets[ggr];
            
            //- SCORING
            if(packet->destX==mv->x && packet->destY==mv->y) score+=packet->weight*100;
        }
    }

    //- Check if packet
    if(getPacketPoint(mv->x, mv->y)>0)
    {
        packetStruct *packet=pfirstGamePacket;

        //- Go through all packets
        while(packet!=NULL)
        {
            if(packet->x==mv->x && packet->y==mv->y)
            {
                //- If we are looking for packets score
                if(robotLookingForPackets!=0)
                {
                    //- SCORING
                    if(packet->score<0) score+=5;
                    else score+=packet->score;
                }
                else
                {
                    //- If not looking for packets but packet is just 5 steps away, score packet
                    //- SCORING
                    if(mv->count<6)
                    {
                        //- Only score if weight is known and we can carry the packet
                        if(packet->weight>0&&packet->weight<=robotCapacity)
                        {
                            if(packet->weight<0) score+=5*100;
                            else score+=packet->weight*100;
                        }
                    }
                }
            }

            packet=packet->next;
        }
    }

    
    if(point==BASE)
    {
        //- SCORING
        //- If looking for packets score base
        if(robotLookingForPackets!=0) score+=100;
        else
        {
            //- If not looking for packets but base is just 5 steps away, score base
            if(mv->count<6) score+=10*100;
        }
    }
    
    //- Add packets
    
    return score;
}

int sumScore(moveStruct *move)
{
    int score;
    int cscore, cscore2;
    int childScore;
    int childNum=0;
    
    if(move==NULL) return 0;
    
    score=move->score;
    
    cscore=0;

    childScore=sumScore(move->children[0]);
    if(childScore>0) { childNum++; cscore+=childScore; }

    childScore=sumScore(move->children[1]);
    if(childScore>0) { childNum++; cscore+=childScore; }

    childScore=sumScore(move->children[2]);
    if(childScore>0) { childNum++; cscore+=childScore; }

    childScore=sumScore(move->children[3]);
    if(childScore>0) { childNum++; cscore+=childScore; }
    
    
    if(childNum<1) childNum=1;
    
    cscore2=(cscore*SCOREHOLD)/(100*childNum);
    if(cscore2<1&&cscore>0) cscore2=1; 
    
    if(cscore2<0) cscore2=0;

    score+=cscore2;
    
    if(score<0) score=0;
    
    move->score=score;
            
    putScorePoint(move->x, move->y, score);
    
    return score;
}

void fortune()
{
    static int SUPER=0;
    
    int ggr;
    moveStruct **tmp;
    int moveScore;

    if(numMoves==0) return;

    //- Go through all moves in moves
    while( numMoves > 0 )
    {
        numMoves--;
        
        
        //- Calc score for this move
        moveScore=score(*move);
        
        (*move)->score=(moveScore>0)?moveScore:0;

        if(moveScore>=SCORE_BREAK)
        {
            int roff=SUPER++;

            //- All four directions
            for(ggr=0;ggr<4;ggr++)
            {
//                int off=(roff+ggr)%4;
                int off=ggr;

                //- Calc new pos
                int x=(*move)->x+moveX[off];
                int y=(*move)->y+moveY[off];
                
                //- If not visited
                if(getMovePoint(x, y)==0)
                {
                    //- Check heap moves
                    if(++numHeapMoves > ROBOT_MAX_MOVES) return;
                    
                    //- Mark this position as visited
                    putMovePoint(x, y, 1);

                    //- Create new move
                    pmoveHeap->x=x;
                    pmoveHeap->y=y;
                    pmoveHeap->count=(*move)->count+1;
                    pmoveHeap->children[0]=NULL;
                    pmoveHeap->children[1]=NULL;
                    pmoveHeap->children[2]=NULL;
                    pmoveHeap->children[3]=NULL;
                    
                    putDistPoint(x, y, pmoveHeap->count);
                    
                    //- Update parent
                    (*move)->children[ggr]=pmoveHeap;
                    
                    //- Add move to next move list
                    *nextMove=pmoveHeap;
                    nextMove++;
                    numNextMoves++;
                    
                    pmoveHeap++;
                }
            }
        }

        move++;
    }
    
    //- Flip pmoves
    tmp=pmoves1;
    pmoves1=pmoves2;
    pmoves2=tmp;
    
    //- Set move pointers
    move=pmoves1;
    nextMove=pmoves2;
    
    //- Set move counters
    numMoves=numNextMoves;
    numNextMoves=0;
    
    fortune();
}

int comparePacketScores(const void *p1, const void *p2)
{
    int score1, score2;

    packetStruct *packet1=*(packetStruct **)p1;
    packetStruct *packet2=*(packetStruct **)p2;
    
    score1=packet1->score;
    score2=packet2->score;

    if(score1<score2) return -1;
    if(score1==score2) return 0;
    return 1;
}

int comparePacketDestinations(const void *p1, const void *p2)
{
    int score1, score2;
    int p1x, p1y, p2x, p2y;

    packetStruct *packet1=*(packetStruct **)p1;
    packetStruct *packet2=*(packetStruct **)p2;


    p1x = (heaviestPacketToPick->destX - packet1->destX);
    p1y = (heaviestPacketToPick->destY - packet1->destY);
    score1 = p1x*p1x + p1y*p1y;

    p2x = (heaviestPacketToPick->destX - packet2->destX);
    p2y = (heaviestPacketToPick->destY - packet2->destY);
    score2 = p2x*p2x + p2y*p2y;
    
    if(score1<score2) return -1;
    if(score1==score2) return 0;
    return 1;
}

void makeMove(int px, int py)
{
    int ggr;
    moveStruct **tmp;
    
    int left, right, up, down, pick, drop;
    int maxscore;
    int weightLeft;
    
    int pickNum;
    int pickIds[10000];
    
    int dropNum;
    int dropIds[10000];
    
    int packetNum;
    packetStruct *packet;
    packetStruct *packets[MAX_GAME_PACKETS];

    #ifdef DEBUG
    printf("Our position=(%d, %d), capacity=%d\n", px, py, robotCapacity);
    #endif
    
    incVisitPoint( px, py );

    // Remove eventual base at current location
    if( BASE == getGamePoint( px, py ) )
        putGamePoint( px, py, PLAIN );
    
    //- Clear move map
    clearScoreMap();
    clearMoveMap();
    putMovePoint(px, py, 1);


    //- Reset counters
    numMoves=0;
    numNextMoves=0;
    numHeapMoves=0;

    //- Reset pointers
    pmoveHeap=&moveHeap[0];
    move=pmoves1;
    nextMove=pmoves2;
    
    //- All four directions
    for(ggr=0;ggr<4;ggr++)
    {
        //- Calc new pos
        int x=px+moveX[ggr];
        int y=py+moveY[ggr];

        //- Create new move
        pmoveHeap->x=x;
        pmoveHeap->y=y;
        pmoveHeap->count=1;
        pmoveHeap->score=0;
        pmoveHeap->children[0]=NULL;
        pmoveHeap->children[1]=NULL;
        pmoveHeap->children[2]=NULL;
        pmoveHeap->children[3]=NULL;
        
        //- Mark this position as visited
        putMovePoint(x, y, 1);

        //- Add move to next move list
        *nextMove=pmoveHeap;
        nextMove++;
        numNextMoves++;

        pmoveHeap++;

        numHeapMoves++;        
    }
    
    //- Flip pmoves
    tmp=pmoves1;
    pmoves1=pmoves2;
    pmoves2=tmp;
    
    //- Set move pointers
    move=pmoves1;
    nextMove=pmoves2;
    
    //- Set move counters
    numMoves=numNextMoves;
    numNextMoves=0;

    #ifdef DEBUG
    printf("My id is %d\n", robotId);
    printf("I am %slooking for packets\n", robotLookingForPackets?"####### ":"****not**** ");
    #endif

    #ifdef DEBUG
    printf("I have %d packets:", robotNumPackets);
    ggr=0;
    for(ggr=0;ggr<robotNumPackets;ggr++)
    {
        printf(" (%d, %d)", robotPackets[ggr]->id, robotPackets[ggr]->weight);
    }
    printf("\n");    
    #endif

    //- Find which packets that are here
    packet=pfirstGamePacket;
    packetNum=0;
    while(packet!=NULL)
    {
        //- Is packet here
        if(packet->x==robotX&&packet->y==robotY)
        {
            packets[packetNum]=packet;
            packetNum++;
        }
        packet=packet->next;
    }

    #ifdef DEBUG
    printf("There is %d packets here\n", packetNum);
    #endif
    
    //- Set packet point
    putPacketPoint(px, py, packetNum);
    
    //- Remove all packets that aren't here but claim to be
    {
        packetStruct *packet=pfirstGamePacket;
        
        while(packet!=NULL)
        {
            //- Does packet claim to be here
            if(packet->x==px && packet->y==py)
            {
                if(packet->timeStamp!=moveCount)
                {
                    removePacket(packet->id);

                    #ifdef DEBUG
                    printf("Removed false packet\n");
                    #endif
                }
            }

            packet=packet->next;
        }
    }    
    
    



    //--------------------------------------------------------------
    //- Calc point score
    fortune();
    


    //--------------------------------------------------------------
    //- Calc packet drop score
    drop=0;

    if(getDestPoint(px, py)!=0)
    {
        //- Find packets to drop
        dropNum=0;
        
        for(ggr=0;ggr<robotNumPackets;ggr++)
        {
            packetStruct *packet=robotPackets[ggr];

            if(packet->destX==px && packet->destY==py)
            {
                dropIds[dropNum++]=packet->id;

                #ifdef DEBUG
                printf("Packet %d is going to be dropped (%d) to (%d, %d)\n", packet->id, packet->weight, packet->destX, packet->destY);
                #endif
                
                drop=1;
            }
        }
        
        dropIds[dropNum]=-1;
    }



    //--------------------------------------------------------------
    //- Calc packet pick score
    pick=0;

    //- Calc score for packets
    for(ggr=0;ggr<packetNum;ggr++)
    {
        int dist, score;    
    
        packetStruct *packet=packets[ggr];
    
        dist=getDistPoint(packet->destX, packet->destY);
        
        //- SCORING
        score=(packet->weight*100)/dist;

        //- SCORING
        score=(score<5)?5:score;

        packet->score=score;

        #ifdef DEBUG
        printf("Packet %d is here (%d of %d) to (%d, %d)\n", packet->id, packet->weight, robotCapacity, packet->destX, packet->destY);
        #endif
    }

    //- Sort packets
    qsort(packets, packetNum, sizeof(void*), comparePacketScores);
    
    #ifdef DEBUG
    printf("qsort done\n");
    #endif

    // Find heaviest package we can carry
    pickNum=0;
    for(ggr=0;ggr<packetNum;ggr++)
    {
        packetStruct *packet=packets[ggr];
        
        if(packet->weight>robotCapacity)
            continue;
        
         // Sort packets after destination distance to heaviest package destination
        heaviestPacketToPick = packet;
        
        #ifdef DEBUG
        printf("heaviestPacketToPick dest = (%d, %d)\n", heaviestPacketToPick->destX, heaviestPacketToPick->destY);
        #endif

        qsort(packets, packetNum, sizeof(void*), comparePacketDestinations);

        break;
    }
    
    
    //- Find packets to pickup
    pickNum=0;
    weightLeft=robotCapacity;
    for(ggr=0;ggr<packetNum;ggr++)
    {
        char info;
        packetStruct *packet=packets[ggr];
        
        if(packet->weight>weightLeft)
            continue;
        
        #ifdef DEBUG
        printf("Packet %d is going to be picked (%d of %d) to (%d, %d)\n", packet->id, packet->weight, robotCapacity, packet->destX, packet->destY);
        #endif

        weightLeft-=packet->weight;
        
        //- Add packet id
        pickIds[pickNum++]=packet->id;
        
        pick=1;
    }
    pickIds[pickNum]=-1;
    
    

    //--------------------------------------------------------------
    //- Calc sum for all moves
    up=sumScore(&moveHeap[0]);
    down=sumScore(&moveHeap[2]);
    left=sumScore(&moveHeap[3]);
    right=sumScore(&moveHeap[1]);
    
    //- Get max score
    maxscore=MAX(MAX(up, down), MAX(left, right));
    
    #ifdef DEBUG
    printf("up=%d, down=%d, left=%d, right=%d, pick=%d, drop=%d, maxscore=%d\n", up, down, left, right, pick, drop, maxscore);
    #endif

    //--------------------------------------------------------------
    if(drop!=0)
    {
        sendDrop(1, dropIds);
    }
    else if(pick!=0)
    {
        sendPick(1, pickIds);
    }
    //- Let last move
    else if(up==maxscore&&lastYMove==UP) sendMove(1, UP);
    else if(down==maxscore&&lastYMove==DOWN) sendMove(1, DOWN);
    else if(left==maxscore&&lastXMove==LEFT) sendMove(1, LEFT);
    else if(right==maxscore&&lastXMove==RIGHT) sendMove(1, RIGHT);
    else if(up==maxscore)
    {
        sendMove(1, UP);
        lastYMove=UP;
    }
    else if(left==maxscore)
    {
        sendMove(1, LEFT);
        lastXMove=LEFT;
    }
    else if(right==maxscore)
    {
        sendMove(1, RIGHT);
        lastXMove=RIGHT;
    }
    else if(down==maxscore)
    {
        sendMove(1, DOWN);
        lastYMove=DOWN;
    }
    

    #ifdef GDEBUG
    DrawScreens();
    #endif
}
